#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.7"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"

###############################################################################
# System imports
from enrichm.genome import Genome, AnnotationParser
from enrichm.gff_generator import GffGenerator
from enrichm.matrix_generator import MatrixGenerator
from enrichm.databases import Databases
from enrichm.sequence_io import SequenceIO
import numpy as np
import statsmodels.sandbox.stats.multicomp as sm
import multiprocessing as mp
import pickle
import shutil
import tempfile
import logging
import subprocess
import os

# Local

###############################################################################
###############################################################################

def parse_genomes(params):
    genome = Genome(*params)
    return genome

###############################################################################
################################ - Classes - ##################################

class Annotate:

    GENOME_BIN                      = 'genome_bin'
    GENOME_PROTEINS                 = 'genome_proteins'    
    GENOME_KO                       = 'annotations_ko'
    GENOME_KO_HMM                   = 'annotations_ko_hmm'
    GENOME_EC                       = 'annotations_ec'
    GENOME_PFAM                     = 'annotations_pfam'
    GENOME_TIGRFAM                  = 'annotations_tigrfam'
    GENOME_HYPOTHETICAL             = 'annotations_hypothetical'
    GENOME_CAZY                     = 'annotations_cazy'
    GENOME_GFF                      = 'annotations_gff'
    GENOME_OBJ                      = 'annotations_genomes'
    OUTPUT_KO                       = 'ko_frequency_table.tsv'
    OUTPUT_KO_HMM                   = 'ko_hmm_frequency_table.tsv'
    OUTPUT_EC                       = 'ec_frequency_table.tsv'
    OUTPUT_PFAM                     = 'pfam_frequency_table.tsv'
    OUTPUT_TIGRFAM                  = 'tigrfam_frequency_table.tsv'
    OUTPUT_CAZY                     = 'cazy_frequency_table.tsv'
    OUTPUT_HYPOTHETICAL_CLUSTER     = 'cluster_frequency_table.tsv'
    OUTPUT_HYPOTHETICAL_ORTHOLOG    = 'ortholog_frequency_table.tsv'
    OUTPUT_HYPOTHETICAL_ANNOTATIONS = 'hypothetical_annotations.tsv'
    OUTPUT_DIAMOND                  = "DIAMOND_search"
    GFF_SUFFIX                      = '.gff'
    PROTEINS_SUFFIX                 = '.faa'
    ANNOTATION_SUFFIX               = '.tsv'
    PICKLE_SUFFIX                   = '.pickle'
    
    def __init__(self,
                 output_directory,
                 ko, ko_hmm, pfam, tigrfam, cluster, ortholog, cazy, ec,
                 evalue, bit, id, aln_query, aln_reference, c, cut_ga, cut_nc, cut_tc, cut_hmm, inflation, chunk_number, chunk_max, count_domains,
                 threads, parallel, suffix, light):

        # Define inputs and outputs
        self.output_directory = output_directory

        # Define type of annotation to be carried out
        self.ko               = ko 
        self.ko_hmm           = ko_hmm
        self.pfam             = pfam 
        self.tigrfam          = tigrfam 
        self.cluster          = cluster 
        self.ortholog         = ortholog

        self.cazy             = cazy
        self.ec               = ec
        
        # Cutoffs
        self.evalue           = evalue 
        self.bit              = bit 
        self.id               = id 
        self.aln_query        = aln_query
        self.aln_reference    = aln_reference
        self.c                = c
        self.cut_ga           = cut_ga
        self.cut_nc           = cut_nc
        self.cut_tc           = cut_tc
        self.cut_hmm          = cut_hmm
        self.inflation        = inflation
        self.chunk_number     = chunk_number
        self.chunk_max        = chunk_max
        self.count_domains    = count_domains

        # Parameters
        self.threads          = threads
        self.parallel         = parallel
        self.suffix           = suffix
        self.light            = light

        # Set up multiprocesses pool
        self.pool             = mp.Pool(processes = int(self.parallel))

        # Load databases
        self.databases        = Databases()

    def prep_genome(self, genome_file_list, genome_directory):
        '''
        Do any preparation specific to the genome annotation pipeline. 
    
        Inputs
        ------
        genome_file_list - List. list of strings, each a path to a file
        containing a genome 

        Outputs
        -------
        returns the directory with all genome ids sym-linked into it.
        '''
        # link all the genomes into one file
        logging.info('Preparing genomes for annotation')
        if genome_file_list:
            os.mkdir(genome_directory)
            genome_paths = list()
            for genome_path in genome_file_list:
                if genome_path.endswith(self.suffix):
                    genome_paths.append(genome_path) 
            cmd = "xargs --arg-file=/dev/stdin cp --target-directory=%s" % genome_directory
            logging.debug(cmd)
            process = subprocess.Popen(["bash", "-c", cmd], 
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       universal_newlines=True)
            process.communicate(input=str('\n'.join(genome_paths)))
        return genome_directory

    def call_proteins(self, genome_directory):
        '''
        Use prodigal (https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-11-119)
        to call proteins within the genomes 
        
        Parameters
        ----------
        genome_directory  - string. Directory containing .fna files for each  
                            input genome

        Outputs
        -------
        returns the directory containing an .faa file for each input genomes
        '''   
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_PROTEINS) 
        os.mkdir(output_directory_path)
        genome_list     = list()
        genome_paths    = list()

        for genome in os.listdir(genome_directory):
            if genome.endswith(self.suffix):
                genome_paths.append(os.path.splitext(genome)[0])

        logging.info("    - Calling proteins for %i genomes" % (len(genome_paths)))
        cmd = "ls %s/*%s | sed 's/%s//g' | grep -o '[^/]*$' | parallel -j %s prodigal -q -p meta -o /dev/null -a %s/{}%s -i %s/{}%s  > /dev/null 2>&1" \
                % (genome_directory,
                   self.suffix,
                   self.suffix,
                   self.parallel,
                   output_directory_path,
                   self.PROTEINS_SUFFIX,
                   genome_directory,
                   self.suffix)

        logging.debug(cmd)
        subprocess.call(cmd, shell = True)
        logging.debug('Finished')
        for genome_protein, genome_nucl in zip(os.listdir(output_directory_path), os.listdir(genome_directory)):
            output_genome_protein_path = os.path.join(output_directory_path, genome_protein)
            output_genome_nucl_path = os.path.join(genome_directory, genome_nucl)
            genome = (self.light, output_genome_protein_path, output_genome_nucl_path)
            genome_list.append(genome)
        
        return genome_list
    
    def annotate_diamond(self, genomes_list, database, parser_type, ids_type, output_subdirectory):
        '''
        Annotate the proteins encoded by each genome with KO ids using either BLAST or using HMM
        searches (no implemented yet).

        Parameters
        ----------        
        genome_faa_directory  - string. Directory containing .faa files for 
                                each input genome
        
        Outputs
        -------
        returns a directory containing the search results for each of the input population genomes, 
        and a frequency matrix contining with the KOs as rows, and the genomes as columns.
        '''        

        output_directory_path = os.path.join(self.output_directory, 
                                             output_subdirectory)
        genome_dict = {genome.name:genome for genome in genomes_list}
        os.mkdir(output_directory_path)
        specific_cutoffs = None
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(str.encode('\n'.join(["sed \"s/>/>%s~/g\" %s" % (genome.name, genome.path) for genome in genomes_list])))
            temp.flush()
            output_annotation_path = os.path.join(output_directory_path, self.OUTPUT_DIAMOND) + self.ANNOTATION_SUFFIX
            logging.info('    - BLASTing genomes' )
            self._diamond_search(temp.name, output_annotation_path, database)

            for genome_name, batch in self.get_batches(output_annotation_path):
                
                if batch:
                    genome = genome_dict[genome_name]
                    genome.add(batch, 
                                 self.evalue, 
                                 self.bit, 
                                 self.aln_query, 
                                 self.aln_reference,
                                 specific_cutoffs,
                                 parser_type,
                               ids_type)
    
    def get_batches(self, input_file):
        last = None
        input_file_io = open(input_file)

        for line in input_file_io:
            split_line = line.strip().split('\t')
            genome_id, _ = split_line[0].split('~')
            
            if last is None:
                last = genome_id
                batch = [split_line]
            
            else:
            
                if last==genome_id:
                    batch.append(split_line)
            
                else:
            
                    yield last, batch
                    batch = [split_line]
                    last = genome_id
        
        if last is None:
        
            yield None, None
        
        else:

            yield last, batch

    def _diamond_search(self, tmp_name, output_path, database):
        '''
        Carry out a diamond blastp search. 

        Parameters
        ----------   
        input_genome_path     - string. Path to file containing .faa file for 
                                an input genome
        output_path           - string. Path to file to output results into    
        databases             - string. Path to HMM to use for searching         
        '''  
         
        cmd = 'bash %s | diamond blastp --quiet --outfmt 6 --max-target-seqs 1 --query /dev/stdin --out %s --db %s --threads %s ' \
                            % (tmp_name, output_path, database, self.threads)
        
        if self.evalue:
            cmd += '--evalue %s ' % (str(self.evalue)) 
        
        if self.bit:
            cmd += '--min-score %s ' % (str(self.bit))
        
        if self.id:
            cmd += '--id %s ' % (str(self.id*100))
        
        if self.aln_query:
            cmd += "--query-cover %s " % (str(self.aln_query * 100))
        
        if self.aln_reference:
            cmd += "--subject-cover %s " % (str(self.aln_reference * 100))

        logging.debug(cmd)
        subprocess.call(cmd, shell = True)
        logging.debug('Finished')

    def hmmsearch_annotation(self, genomes_list, output_directory_path, database, ids_type, parser):
        '''
        Annotate the proteins encoded by each genome with pfam ids using HMM searches.

        Parameters
        ----------        
        genomes_list  - list. list of Genome objects        

        '''    
        os.mkdir(output_directory_path)
        genome_dict = {genome.name: genome for genome in genomes_list}

        if (ids_type == AnnotationParser.TIGRFAM or 
            ids_type == AnnotationParser.PFAM):
            hmmcutoff=True
        else:
            hmmcutoff=False
        
        if ids_type == AnnotationParser.KO_HMM:
            specific_cutoffs = self.databases.parse_ko_cutoffs()
        else:
            specific_cutoffs = None

        self._hmm_search(output_directory_path, database, hmmcutoff)

        for genome_annotation in os.listdir(output_directory_path):
            genome_id = os.path.splitext(genome_annotation)[0]
            genome = genome_dict[genome_id]
            output_annotation_path = os.path.join(output_directory_path, genome_annotation)
            genome.add(output_annotation_path, 
                         self.evalue, 
                         self.bit, 
                         self.aln_query, 
                         self.aln_reference,
                         specific_cutoffs,
                         parser,
                         ids_type)

    def annotate_hypothetical(self, genomes_list, ortholog):
        '''
        Sort proteins coded by each genome into homologous clusters.  
        
        Inputs
        ------
        genomes_list  - list. list of Genome objects        

        '''
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_HYPOTHETICAL)
        os.mkdir(output_directory_path)      

        with tempfile.NamedTemporaryFile() as temp:

            temp.write(str.encode('\n'.join(["sed \"s/>/>%s~/g\" %s" % (genome.name, genome.path) for genome in genomes_list])))
            temp.flush()  

            tmp_dir = tempfile.mkdtemp()  
                
            db_path = os.path.join(output_directory_path, "db")
            clu_path = os.path.join(output_directory_path, "clu")
            align_path = os.path.join(output_directory_path, "alignDb")
            blast_output_path = os.path.join(output_directory_path, "alignDb.m8")
            formatted_blast_output_path = os.path.join(output_directory_path, "alignDb.formatted.m8")

            clu_tsv_path = os.path.join(output_directory_path, "hypothetical_clusters.tsv")
            logging.info('    - Generating MMSeqs2 database')
            cmd = "bash %s | sponge | mmseqs createdb /dev/stdin %s -v 0 > /dev/null 2>&1" % (
                temp.name, db_path)
            logging.debug(cmd)
            subprocess.call(cmd, shell = True)
            logging.debug('Finished')
            logging.info('    - Clustering genome proteins')
            cmd = "mmseqs cluster %s %s %s --max-seqs 1000 --threads %s --min-seq-id %s -e %f -c %s -v 0 " % (
                db_path, clu_path, tmp_dir, self.threads, self.id, self.evalue, self.c)
            logging.debug(cmd)
            subprocess.call(cmd, shell = True)
            logging.debug('Finished')
            logging.info('    - Extracting clusters')

            cmd = 'mmseqs createtsv %s %s %s %s -v 0 > /dev/null 2>&1' % (
                db_path, db_path, clu_path, clu_tsv_path)
            logging.debug(cmd)
            subprocess.call(cmd, shell = True)
            logging.debug('Finished')
            
            logging.info('    - Computing Smith-Waterman alignments for clustering results')
            cmd = "mmseqs alignall %s %s %s --alignment-mode 3 -v 0  " % (
                db_path, clu_path, align_path)
            logging.debug(cmd)
            subprocess.call(cmd, shell = True)
            logging.debug('Finished')

            logging.info('    - Converting to BLAST-like output')
            cmd = "mmseqs createtsv %s %s %s %s -v 0 > /dev/null 2>&1   " % (
                db_path, db_path, align_path, blast_output_path)
            # --format-output query,target,bits
            logging.debug(cmd)
            subprocess.call(cmd, shell = True)
            logging.debug('Finished')
            logging.info('    - Reformatting BLAST output')
            cmd = "OFS=\"\t\" awk 'FNR==NR{a[$1]=$2;next}{$3=a[$3]; $1=\"\"; for(i=2;i<NF;i++){printf(\"%s\t\",$i)} printf(\"\\n\")}' %s %s | cut -f1,2,5 > %s" % ("%s", db_path + '.lookup', blast_output_path, formatted_blast_output_path)
            logging.debug(cmd)
            subprocess.call(cmd, shell=True)
            logging.debug('Finished')

        ortholog_dict = self.run_mcl(formatted_blast_output_path,
                                     output_directory_path)
        ortholog_ids = ortholog_dict.keys()
        cluster_ids = self.parse_cluster_results(clu_tsv_path,
                                        genomes_list,
                                                ortholog_dict,
                                        output_directory_path)
        return cluster_ids, ortholog_ids

    def run_mcl(self, blast_abc, output_directory_path):
        dict_path = os.path.join(output_directory_path, "alignDb.dict")
        mci_path = os.path.join(output_directory_path, "alignDb.mci")
        cluster_path = os.path.join(output_directory_path, "mcl_clusters.tsv")
        output_path = os.path.join(output_directory_path, "mcl_clusters.convert.tsv")

        logging.info('    - Preparing network')
        ortholog_dict = dict()
        cmd = "mcxload -abc %s -write-tab %s -o %s --stream-mirror --stream-neg-log10 > /dev/null 2>&1" % (
            blast_abc, dict_path, mci_path)
        logging.debug(cmd)
        subprocess.call(cmd, shell=True)
        logging.debug('Finished')

        logging.info('    - Finding orthologs')
        ortholog_dict = dict()
        cmd = 'mcl %s -te %s -I %s -o %s  > /dev/null 2>&1' % (
            mci_path, self.threads, self.inflation, cluster_path)
        logging.debug(cmd)
        subprocess.call(cmd, shell = True)
        logging.debug('Finished')
        
        logging.info('    - Reformatting output')
        ortholog_dict = dict()
        cmd = 'mcxdump -icl %s -o %s -tabr %s > /dev/null 2>&1' % (cluster_path,  output_path, dict_path)
        logging.debug(cmd)
        subprocess.call(cmd, shell=True)
        logging.debug('Finished')

        ortholog = 1
        for line in open(output_path):
            ortholog_idx = "ortholog_%i" % ortholog
            ortholog_dict[ortholog_idx] = set()
        
            for protein in line.strip().split('\t'):
                ortholog_dict[ortholog_idx].add(protein)
        
            ortholog += 1
        
        return ortholog_dict

    def parse_cluster_results(self, 
                              cluster_output_path,
                              genomes_list,
                              ortholog_dict,
                              output_directory_path):
        '''
        Parse cluster output in tab format.
        
        Inputs
        ------
        from_cluster_results    - String. Path to mmseqs2 clustering output file
        
        Yields
        -------
        A cluster name, and a list of sequences in that cluster.
        
        '''
        logging.info('    - Parsing input cluster file: %s' % cluster_output_path)
        
        cluster_ids             = set()
        previous_cluster_name   = None
        counter                 = 0
        genome_dictionary       = {genome.name:genome for genome in genomes_list}

        with open(os.path.join(output_directory_path, self.OUTPUT_HYPOTHETICAL_ANNOTATIONS), 'w') as out_io:

            for line in open(cluster_output_path):
    
                cluster_id, member      = line.strip().split('\t')
                genome_id, sequence_id  = member.split('~')
                
                if cluster_id == previous_cluster_name:
                    genome_dictionary[genome_id].add_cluster(sequence_id, "cluster_%i" % counter)
                else:
                    counter += 1
                    previous_cluster_name = cluster_id 
                    cluster_ids.add("cluster_%i" % counter)
                    genome_dictionary[genome_id].add_cluster(sequence_id, "cluster_%i" % counter)
                out_io.write('\t'.join([genome_id, sequence_id, "cluster_%i" % counter]) + '\n')
        
        for ortholog, group in ortholog_dict.items():
            for member in group:
                genome, protein = member.split('~')
                genome_dictionary[genome].add_ortholog(protein, ortholog)
        
        return cluster_ids

    def _default_hmmsearch_options(self):
        cmd = ''
        if self.bit:
            cmd += '-T %s ' % (str(self.bit))    
        else:
            cmd += '-E %s ' % (str(self.evalue)) 
        return cmd

    def _hmm_search(self, output_path, database, hmmcutoff):
        '''
        Carry out a hmmsearch. 

        Parameters
        ----------   
        input_genome_path     - string. Path to file containing .faa file for 
                                an input genome
        output_path           - string. Path to file to output results into   
        databases             - string. Path to HMM to use for searching          
        '''
        
        input_genome_path = os.path.join(self.output_directory, self.GENOME_PROTEINS)
        cmd = "ls %s | sed 's/%s//g' | parallel -j %s hmmsearch --cpu %s -o /dev/null --noali --domtblout %s/{}%s " \
                          % (input_genome_path, self.PROTEINS_SUFFIX, self.parallel, 
                             self.threads, output_path, self.ANNOTATION_SUFFIX)
        if hmmcutoff:
            if(self.cut_ga or self.cut_nc or self.cut_tc):

                if self.cut_ga:
                    cmd += " --cut_ga "
                if self.cut_nc:
                    cmd += " --cut_nc "
                if self.cut_tc:
                    cmd += " --cut_tc "
            else:
                cmd += self._default_hmmsearch_options()
        else:
            cmd += self._default_hmmsearch_options()   

        cmd += "%s %s/{}.faa 2> /dev/null" % (database, input_genome_path)

        logging.debug(cmd)
        subprocess.call(cmd, shell = True)        
        logging.debug('Finished')
    
    def _generate_gff_files(self, genomes_list):
        '''
        Write GFF files for each of the genome objects in genomes_list

        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_GFF)
        os.mkdir(output_directory_path)      
        for genome in genomes_list:
            logging.info('    - Generating .gff file for %s' % genome.name)
            gff_output = os.path.join(output_directory_path, genome.name + self.GFF_SUFFIX)
            gg = GffGenerator()
            gg.write(genome, gff_output)

    def _rename_fasta(self, genomes_list):
        '''
        Rename the called proteins with annotation ids.
        
        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        seqio = SequenceIO()
        for genome in genomes_list:
            fd, fname = tempfile.mkstemp(suffix='.faa', text=True)
            with open(fname, 'w') as out_io:
                for description, sequence in seqio.each(open(genome.path)):
                    name = description.partition(' ')[0]
                    annotations = ' '.join(genome.sequences[name].all_annotations())
                    out_io.write( ">%s %s\n" % (name, annotations) )
                    out_io.write( str(sequence) + '\n' )
            os.close(fd)
            logging.debug('Moving %s to %s' % (fname, genome.path))
            shutil.move(fname, genome.path)

    def _pickle_objects(self, genomes_list):
        '''
        Store annotated genome objects as pickles.
        
        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = os.path.join(self.output_directory,
                                             self.GENOME_OBJ)
        os.mkdir(output_directory_path)
        for genome in genomes_list:
            with open(os.path.join(output_directory_path, genome.name + self.PICKLE_SUFFIX), 'wb') as output:
                pickle.dump(genome, output)

    def list_splitter(self, input_list, chunk_number, chunk_max):
        list_size   = float( len(input_list) )
        chunk_size  = int(round( (list_size/chunk_number), 0 ))


        if chunk_size > chunk_max:
            chunk_size = chunk_max
        elif chunk_size < 1:
            chunk_size = list_size

        while list_size > 0:

            if len(input_list)<=chunk_size:
                yield input_list
                del input_list
            else:
                yield input_list[:chunk_size]
                del input_list[:chunk_size]

            try:
                list_size = len(input_list)
            except NameError:
                list_size = 0

    def parse_genome_inputs(self, genome_directory, protein_directory, genome_files, protein_files):
        '''
        Inputs
        ------
        
        Outputs
        -------
        
        '''

        prep_genomes_list   = list()  
        genomes_list        = list()

        if protein_directory:
            logging.info("Using provided proteins")
            protein_genome_list = list()
            
            for protein_file in os.listdir(protein_directory):
                protein_genome_list.append(os.path.join(protein_directory, protein_file))

            directory = self.prep_genome(protein_genome_list,
                                         os.path.join(self.output_directory,
                                                      self.GENOME_PROTEINS))
            
            for genome_proteins_file in os.listdir(directory):

                if genome_proteins_file.endswith(self.suffix):
                    genome = (self.light, os.path.join(directory, genome_proteins_file), None)
                    prep_genomes_list.append(genome)

        elif protein_files:
            logging.info("Using provided proteins")
            directory = self.prep_genome(protein_files,
                             os.path.join(self.output_directory,
                                          self.GENOME_PROTEINS))
            
            for protein_file in os.listdir(directory):
                prep_genomes_list.append((self.light, os.path.join(directory, os.path.basename(protein_file)), None))

        elif genome_directory:
            logging.info("Calling proteins for annotation")
            prep_genomes_list = self.call_proteins(genome_directory)
            directory = genome_directory
        
        elif genome_files:
            logging.info("Calling proteins for annotation")      
            directory = self.prep_genome(genome_files,
                                         os.path.join(self.output_directory,
                                                      self.GENOME_BIN))      
            prep_genomes_list = self.call_proteins(directory)
        
        
        for chunk in self.list_splitter(prep_genomes_list, self.chunk_number, self.chunk_max):
            genomes_list += self.pool.map(parse_genomes, chunk)

        return genomes_list

    def do(self, genome_directory, protein_directory, genome_files, protein_files):
        '''
        Run Annotate pipeline for enrichM

        Parameters
        ----------
        genome_directory    - String. Path to directory containing genomes
        protein_directory   - String. Path to directory containing proteins (.faa files) for genomes
        genome_files        - List. List of strings, each to a .fna genome file.
        protein_files       - List. List of strings, each to a .faa proteins file.
        '''

        logging.info("Running pipeline: annotate")
        logging.info("Setting up for genome annotation")
        genomes_list = self.parse_genome_inputs(genome_directory, protein_directory, genome_files, protein_files)
        if len(genomes_list)==0:
            logging.error('There were no genomes found with the suffix %s within the provided directory' \
                                        %  (self.suffix))
        else:
            logging.info("Starting annotation:")

            if (self.cluster or self.ortholog):
                logging.info('    - Annotating genomes with hypothetical clusters')
                cluster_ids, ortholog_ids = self.annotate_hypothetical(genomes_list, self.ortholog)

                logging.info('    - Generating hypotheticals frequency table') 
                mg = MatrixGenerator(MatrixGenerator.HYPOTHETICAL, cluster_ids)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_HYPOTHETICAL_CLUSTER)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

                if self.ortholog:
                    mg = MatrixGenerator(MatrixGenerator.ORTHOLOG, ortholog_ids)
                    freq_table = os.path.join(self.output_directory, self.OUTPUT_HYPOTHETICAL_ORTHOLOG)
                    mg.write_matrix(genomes_list, self.count_domains, freq_table)
                

            if self.ko:
                annotation_type = AnnotationParser.BLASTPARSER
                logging.info('    - Annotating genomes with ko ids using DIAMOND')
                self.annotate_diamond(
                    genomes_list, self.databases.KO_DB, annotation_type, AnnotationParser.KO, self.GENOME_KO)

                logging.info('    - Generating ko frequency table')
                mg = MatrixGenerator(MatrixGenerator.KO)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_KO)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.ko_hmm:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with ko ids using HMMs')
                self.hmmsearch_annotation(genomes_list,
                                          os.path.join(
                                              self.output_directory, self.GENOME_KO_HMM),
                                          self.databases.KO_HMM_DB,
                                          AnnotationParser.KO,
                                          annotation_type)

                logging.info('    - Generating ko frequency table')
                mg = MatrixGenerator(MatrixGenerator.KO)
                freq_table = os.path.join(
                    self.output_directory, self.OUTPUT_KO_HMM)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.ec:
                annotation_type = AnnotationParser.BLASTPARSER
                logging.info('    - Annotating genomes with ec ids')
                self.annotate_diamond(genomes_list, self.databases.EC_DB, annotation_type, AnnotationParser.EC, self.GENOME_EC)
                
                logging.info('    - Generating ec frequency table')
                mg = MatrixGenerator(MatrixGenerator.EC)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_EC)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.pfam:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with pfam ids')
                self.hmmsearch_annotation(genomes_list,
                                          os.path.join(self.output_directory, self.GENOME_PFAM),
                                          self.databases.PFAM_DB,
                                          AnnotationParser.PFAM,
                                          annotation_type)
                
                logging.info('    - Generating pfam frequency table')
                mg = MatrixGenerator(MatrixGenerator.PFAM)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_PFAM)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.tigrfam:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with tigrfam ids')
                self.hmmsearch_annotation(genomes_list,
                                          os.path.join(self.output_directory, self.GENOME_TIGRFAM),
                                          self.databases.TIGRFAM_DB,
                                          AnnotationParser.TIGRFAM,
                                          annotation_type)

                logging.info('    - Generating tigrfam frequency table')
                mg = MatrixGenerator(MatrixGenerator.TIGRFAM)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_TIGRFAM)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)
            
            if self.cazy:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with CAZY ids')
                self.hmmsearch_annotation(genomes_list,
                                          os.path.join(self.output_directory, self.GENOME_CAZY),
                                          self.databases.CAZY_DB,
                                          AnnotationParser.CAZY,
                                          annotation_type)

                logging.info('    - Generating CAZY frequency table')
                mg = MatrixGenerator(MatrixGenerator.CAZY)
                freq_table = os.path.join(self.output_directory, self.OUTPUT_CAZY)
                mg.write_matrix(genomes_list, self.count_domains, freq_table)

            if hasattr(list(genomes_list[0].sequences.values())[0], "prod_id"):
                logging.info('Generating .gff files:')
                self._generate_gff_files(genomes_list)

                logging.info('Renaming protein headers')
                self._rename_fasta(genomes_list)

            if not self.light:
                logging.info('Storing genome objects')
                self._pickle_objects(genomes_list)

            logging.info('Finished annotation')




#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"
 
###############################################################################
# System imports

class GffGenerator():

	def write(self, genome, output_file):
		'''
		Writes a gff file for a Genome object
		
		Parameters
		----------
		genome 			- Genome object
		output_file 	- string. file name to output results to.
		'''

		with open(output_file, 'w') as out_io:
			for sequence in genome.ordered_sequences():
				contig_id = '_'.join(sequence.seqname.split('_')[:-1])
				
				features  = ['seq_id=%s'		% sequence.seqname,
						 	 'prodigal_id=%s' 	% sequence.prod_id,
							 'partial=%s' 	 	% sequence.partial,
							 'start_type=%s'  	% sequence.starttype,
							 'rbs_motif=%s'   	% sequence.rbs_motif,
							 'rbs_spacer=%s'  	% sequence.rbs_spacer,
							 'gc=%s'    		% sequence.gc]

				if len(sequence.all_annotations())>0:
					features.append('annotations=%s' % ','.join(sequence.all_annotations()))
				else:
					features.append('annotations=hypothetical_protein')

				line = [contig_id, 
						'prodigal',
						'CDS',
						sequence.startpos,
						sequence.finishpos,
						'.',
						('-' if sequence.direction == '-1'
						 else '+'),
						'0',
						';'.join(features)]
						
				out_io.write('\t'.join(line) + '\n')


import sys
import re
import pickle
KEGG = '^(K\d{5})'

def ex(string):
    chunk_split = chunk.split('-')

output = sys.argv[2]
output_dict = dict()
for line in open(sys.argv[1]):
    module, description = line.strip().split('\t')
    chunks = []
    for chunk in description.split(' '):

        if '-' in chunk:
            if chunk == '--':continue
            to_keep = []

            while '-' in chunk:
                chunk_split = chunk.split('-')
                to_keep.append(chunk_split[0])
                to_assess = '-'.join(chunk_split[1:])

                if re.match(KEGG, to_assess):
                    chunk = chunk.replace('-' + to_assess[:6], "")

                else:

                    if to_assess.startswith('('):
                        chunk=chunk.replace('-' + to_assess[0:to_assess.index(')')+1], "")
        chunks.append(chunk)
    description = ' '.join(chunks) 
    if '(' in description:
        inside = False
        for i in description:
            if inside:
                count += i
            if i == '(':
                count = ''
                inside = True
            elif i == ')':

                if len(count[:-1]) == 6:
                    description= description.replace(
                        "(%s)" % count[:-1], count[:-1])
                inside = False
                count = ''

    while '  ' in description:
        description= description.replace("  ", " ")
    while description.endswith(' '):
        description = description[:-1]
    output_dict[module] = description

out_io = open(output, 'wb')
pickle.dump(output_dict, out_io)



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2018"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import pickle
import os
import logging
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.model_selection import train_test_split, RandomizedSearchCV, GridSearchCV
# Local
import enrichm.generate
################################################################################

class GenerateModel():

    def __init__(self):
        '''     
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        # Subparser names
        self.REGRESSOR = "regressor"
        self.CLASSIFIER = "classifier"

        # Output file names
        self.ATTRIBUTE_IMPORTANCES = 'attribute_importances.tsv'
        self.ATTRIBUTE_LIST = "attribute_list.txt"
        self.MODEL_PICKLE = "rf_model.pickle"
        self.LABELS_DICT = "labels_dict.pickle"
    

    
    def parse_input_matrix(self, input_matrix_path):
        '''     
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        input_matrix_io = open(input_matrix_path)
        headers = input_matrix_io.readline().strip().split('\t')[1:]
        
        output_dictionary = {header:{} for header in headers}
        attribute_list = list()

        for line in input_matrix_io:
            values = line.strip().split('\t')
            attribute_value = values[0]
            attribute_list.append(attribute_value)
        
            for header, value in zip(headers, values[1:]):
                output_dictionary[header][attribute_value] = value
        logging.info('\t\tRead in matrix of %i samples associated with %i attributes' \
                % (len(output_dictionary), len(attribute_list)))
        return output_dictionary, attribute_list

    def parse_groups_matrix(self, groups_matrix_path):
        '''     
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        output_dictionary = dict()

        for line in open(groups_matrix_path):
            attribute_value, group = line.strip().split('\t')

            if group in output_dictionary:
                raise Exception("Duplicated entry in groups file: %s" % attribute_value)
            else:
                output_dictionary[attribute_value] = group

        logging.info('\t\tRead in metadata for %i samples' % (len(output_dictionary)))
        
        return output_dictionary

    def _numerify(self, input_list):
        '''
        
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        idx = 0
        output_dictionary = {}
        output_list = []
        
        for group in input_list:
            if group not in output_dictionary:
                output_dictionary[group] = idx
                idx += 1
            output_list.append(output_dictionary[group])
        output_dictionary = {item:key for key, item in output_dictionary.items()}
        return output_dictionary, output_list

    def _write_attribute_list(self, attribute_list, output_directory):
        with open(os.path.join(output_directory, self.ATTRIBUTE_LIST), 'wb') as out_io:
            out_io.write(str.encode('\n'.join(attribute_list)))
    
    def _write_importances(self, model, attribute_list, output_directory):
        '''
        
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        importances = list(model.feature_importances_)
        # List of tuples with variable and importance
        feature_importances = [(feature, round(importance, 2)) for feature, importance in zip(attribute_list, importances)]
        # Sort the feature importances by most important first
        feature_importances = sorted(feature_importances, key = lambda x: x[1], reverse = True)
        # Print out the feature and importances 
        logging.info('%i attributes found with an importance > 0' % (len([x for x in feature_importances if x[1]>0])))
        logging.info('Writing attribute importances')
        with open(os.path.join(output_directory, self.ATTRIBUTE_IMPORTANCES), 'wb') as out_io:
            out_io.write(str.encode('\t'.join(['Variable', 'Importance']) + '\n'))
            for pair in feature_importances:
                var, imp = pair
                out_io.write(str.encode('\t'.join([str(var), str(imp)]) + '\n'))
    
    def transpose(self, labels, features, attribute_list):
        '''
        
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        labels_list     = list()
        features_list   = list()

        for col in list(labels.keys()):

            labels_list.append(labels[col])
            col_values = list()

            for attribute in attribute_list:
                col_values.append(features[col][attribute])
            
            features_list.append(col_values)

        return labels_list, features_list

    def grid_search_cv(self, random_search_cv, threads, rf):
        '''
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        logging.info('Generating model for grid search cross validation')

        bootstrap = [random_search_cv.best_params_['bootstrap']]
        max_depth = range(random_search_cv.best_params_['max_depth']-9, random_search_cv.best_params_['max_depth']+10, 2)
        max_features = [random_search_cv.best_params_['max_features'], 2, 3]
        min_samples_leaf = range(random_search_cv.best_params_['min_samples_leaf']-1, random_search_cv.best_params_['min_samples_leaf']+2)
        min_samples_split = range(random_search_cv.best_params_['min_samples_split']-2, random_search_cv.best_params_['min_samples_split']+3, 2)
        n_estimators = range(random_search_cv.best_params_['n_estimators']-200, random_search_cv.best_params_['n_estimators'] + 200, 100)

        param_grid = {
            'bootstrap':bootstrap,
            'max_depth':[x for x in max_depth if x>0],
            'max_features':max_features,
            'min_samples_leaf':[x for x in min_samples_leaf if x>0],
            'min_samples_split':[x for x in min_samples_split if x>0],
            'n_estimators':[x for x in n_estimators if x>0]
        }
        

        grid_search = GridSearchCV(estimator = rf,
                                   param_grid = param_grid, 
                                   cv = 3,
                                   n_jobs = threads)
        return grid_search

    def random_search_cv(self, threads, rf):
        '''
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        logging.info('Generating model for random search cross validation')
        
        n_estimators            = [int(x) for x in np.linspace(start = 200, stop = 2000, num = 10)]
        max_features            = ['auto', 'sqrt']
        max_depth               = [int(x) for x in np.linspace(10, 110, num = 11)]
        max_depth.append(None)
        min_samples_split       = [2, 5, 10]
        min_samples_leaf        = [1, 2, 4]
        bootstrap               = [True, False]
        random_grid = {'n_estimators': n_estimators,
                       'max_features': max_features,
                       'max_depth': max_depth,
                       'min_samples_split': min_samples_split,
                       'min_samples_leaf': min_samples_leaf,
                       'bootstrap': bootstrap}

        rf_random = RandomizedSearchCV(estimator = rf,
                                        param_distributions = random_grid,
                                        n_iter = 100,
                                        cv = 3,
                                        random_state = 42,
                                        n_jobs = threads)
        return rf_random

    def tune(self, features_list, labels_list_numeric, testing_portion, grid_search, threads, rf):
        '''     
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        logging.info('Splitting data into test and training datasets')
        train_features, test_features, train_labels, test_labels \
            = train_test_split(features_list,
                               labels_list_numeric,
                               test_size = testing_portion,
                               random_state = 7)


        rf_random_model = self.random_search_cv(threads, rf)
        logging.info('Fitting model')
        rf_random_trained_model = rf_random_model.fit(train_features, train_labels)

        logging.info('Best parameters from random search cross validation:')
        for x,y in rf_random_trained_model.best_params_.items():
            logging.info("\t\t%s: %s" % (x, str(y)))
        if grid_search:
            rf_grid_model = self.grid_search_cv(rf_random_model, threads, rf)   
            logging.info('Fitting model')
            rf_grid_trained_model = rf_grid_model.fit(train_features, train_labels)

            logging.info('Best parameters from grid search cross validation:')
            for x,y in rf_grid_trained_model.best_params_.items():
                logging.info("\t\t%s: %s" % (x, str(y)))

            best_model = rf_grid_trained_model.best_estimator_
        else:
            best_model = rf_random_trained_model.best_estimator_

        return best_model, test_features, test_labels

    def do(self, input_matrix_path, groups_path, model_type,
           testing_portion, grid_search, threads, output_directory):
        '''
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        logging.info('Using %f%% of the input data for testing' % (testing_portion*100))

        if model_type == self.REGRESSOR:
            model = RandomForestRegressor()
        elif model_type == self.CLASSIFIER:
            model = RandomForestClassifier()
        else:
            raise Exception("Model type not recognised: %s" % (model_type))

        logging.info('Parsing inputs:')
        labels \
            = self.parse_groups_matrix(groups_path)
        features, attribute_list \
            = self.parse_input_matrix(input_matrix_path)

        labels_list, features_list = self.transpose(labels, features, attribute_list)

        labels_dict, labels_list_numeric = self._numerify(labels_list)
        labels_list_numeric     =   np.array(labels_list_numeric)
        features_list   =   np.array(features_list)
        logging.info('Splitting data into training and testing portions')

        logging.info("Tuning hyperparameters")
        rf, test_features, test_labels = self.tune(features_list,
                                                   labels_list_numeric,
                                                   testing_portion,
                                                   grid_search,
                                                   threads,
                                                   model)

        logging.info('Making predictions on test data:')
        predictions = rf.predict(test_features)
        errors = abs(predictions - test_labels)
        logging.info('\t\tMean Absolute Error: %f degrees' % (round(np.mean(errors), 2)))
        
        correctness = []
        for prediction, label in zip(np.round(predictions), test_labels):
            if prediction==label:
                correctness.append(1)
            else:
                correctness.append(0)

        accuracy = (sum(correctness)/float(len(correctness)))*100
        logging.info('\t\tAccuracy: %f%%' % (round(accuracy, 2)))
        self._write_importances(rf, attribute_list, output_directory)

        self._write_attribute_list(attribute_list, output_directory)

        logging.info("Preserving model")
        pickle.dump(rf, open(os.path.join(output_directory, self.MODEL_PICKLE) , 'wb'))

        logging.info("Preserving group labels")
        pickle.dump(labels_dict, open(os.path.join(output_directory, self.LABELS_DICT) , 'wb'))



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Ben Woodcroft"
__copyright__ = "Copyright 2017"
__credits__ = ["Ben Woodcroft"]
__license__ = "GPL3+"
__maintainer__ = "Ben Woodcroft"
__email__ = "b.woodcroft near uq.edu.au"
__status__ = "Development"

###############################################################################
# Imports
import os
import re
import logging
from itertools import chain
# Local
from enrichm.databases import Databases
from enrichm.module_description_parser import ModuleDescription
from enrichm.parser import Parser
###############################################################################

class Classify:
    
    KO_OUTPUT       = "module_completeness.tsv"
    MODULE_PATHS    = "module_paths.tsv"
    AGGREGATE_OUTPUT = "aggregate_output.tsv"

    def __init__(self):

        d=Databases()
        self.ko_re              = re.compile('^K\d+$')
        self.signature_modules  = d.signature_modules
        self.m2def              = d.m2def
        self.m                  = d.m

    def _update_with_custom_modules(self, custom_modules):
        custom_modules_dict = dict()
        for line in open(custom_modules):
                custom_modules_dict[line.split('\t')[0]] = line.strip().split('\t')[1]
        self.m2def.update(custom_modules_dict)
        
        for key in custom_modules_dict.keys():
            self.m[key] = 'Custom'
    
    def write(self, lines, output_path):
        
        logging.info('Writing results to file: %s' % output_path)

        with open(output_path, 'w') as output_path_io:
            for line in lines:
                output_path_io.write(line)

    def do(self, custom_modules, cutoff, aggregate, genome_and_annotation_file, 
           genome_and_annotation_matrix, output_directory):
        '''
        
        Parameters
        ----------
        custom_modules                  - string. Path to file containing custom module definitions, 
                                          consistent with KEGG module nomenclature 
                                          (http://www.genome.jp/kegg/module.html)
        cutoff                          - float. Fraction of a module needed in order to be included
                                          in the output.
        genome_and_annotation_file      - string. Path to file containing genome - annotation file. This file 
                                          contains two columns, the first with the genome name, the 
                                          second with a annotation annotation within that genome 
        genome_and_annotation_matrix    - string. Path to file containing genome - annotation matrix
        output                          - string. Path to file to output results to.

        '''
        
        pathway             = dict()
        genome_output_lines = list()

        if custom_modules:
            logging.info('Reading in custom modules: %s' % custom_modules)
            self._update_with_custom_modules(custom_modules)
        
        if genome_and_annotation_file:
            genome_to_annotation_sets = Parser.parse_genome_and_annotation_file_lf(genome_and_annotation_file)
        
        elif genome_and_annotation_matrix:
            genome_to_annotation_sets = Parser.parse_genome_and_annotation_file_matrix(genome_and_annotation_matrix)

        if aggregate:
            logging.info('Reading in abundances: %s' %
                         (genome_and_annotation_matrix))
            abundances = Parser.parse_simple_matrix(genome_and_annotation_matrix, True)
            abundance_result = dict()

            
        logging.info("Read in annotations for %i genomes" % len(genome_to_annotation_sets))
        
        output_lines = ['\t'.join(["Genome_name", "Module_id", "Module_name", "Steps_found", 
                             "Steps_needed", "Percent_steps_found"]) + '\n'] # "KO_found", "KO_needed", "Percent_KO_found"
        
        genome_output_lines = ['\t'.join(["Genome_name", "Module_id", "Module_name"]) + '\n']

        for name, pathway_string in self.m2def.items():
        
            if name not in self.signature_modules:   
                path = ModuleDescription(pathway_string)
                pathway[name] = path

                for genome, annotations in genome_to_annotation_sets.items():    

                    num_covered, _, _, ko_path = path.num_covered_steps(annotations)
                    num_all         = path.num_steps()
                    perc_covered    = num_covered / float(num_all)
                    ko_path_list    = list(chain(*ko_path.values()))

                    if perc_covered >= cutoff:

                        if path.is_single_step:
                        
                            if perc_covered!=1:

                                if cutoff < 1:
                                    num_all = 1
                                    num_covered = 0
                                    perc_covered = 0.0

                                else:
                                    continue
                            
                            else:
                                num_all = 1
                                num_covered = 1

                        if aggregate:
                            
                            if genome not in abundance_result:
                                abundance_result[genome] = dict()
                            pathway_abundance = [abundances[genome][ko] for ko in ko_path_list]
                            pathway_average_abundance = sum(pathway_abundance) / len(pathway_abundance)
                            abundance_result[genome][name] = pathway_average_abundance

                        genome_output_lines.append('\t'.join([genome, name, self.m[name], 
                                                              ','.join(ko_path_list) + '\n']))
                        output_line = "\t".join([genome, name, self.m[name],
                                                  str(num_covered), 
                                                  str(num_all),
                                                  str(round(perc_covered * 100, 2))
                                                  ]) 
                        output_lines.append(output_line + '\n') 
                        
        self.write(output_lines, os.path.join(output_directory, self.KO_OUTPUT))
        self.write(genome_output_lines, os.path.join(output_directory, self.MODULE_PATHS)) 
        
        if aggregate:
            samples = list(abundance_result.keys() )
            output_lines = ['\t'.join(["ID"] + samples) + '\n']
            for module in self.m2def.keys():
            
                if module not in self.signature_modules:
                    ol = [module]

                    for sample in samples:
            
                        if module in abundance_result[sample]:
                            ol.append(str(abundance_result[sample][module]))
            
                        else:
                            ol.append('0.0')
                    output_lines.append('\t'.join(ol) + '\n')
            self.write(output_lines, os.path.join(output_directory, self.AGGREGATE_OUTPUT))






#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2018"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import logging
import sys
import os
import shutil
import time

# Local
from enrichm.data import Data
from enrichm.network_analyzer import NetworkAnalyser
from enrichm.enrichment import Enrichment
from enrichm.annotate import Annotate
from enrichm.classifier import Classify
from enrichm.generate import GenerateModel
from enrichm.predict import Predict
from enrichm.connect import Connect
from enrichm.aggregate import Aggregate
###############################################################################

debug={1:logging.CRITICAL,
       2:logging.ERROR,
       3:logging.WARNING,
       4:logging.INFO,
       5:logging.DEBUG}

###############################################################################

class Run:
    
    def __init__(self):
        
        self.ANNOTATE        = 'annotate'
        self.COMPARE         = 'compare'

        self.CLASSIFY        = 'classify'
        self.BUILD           = 'build'
        self.ENRICHMENT      = 'enrichment'
        self.MODULE_AB       = 'module_ab'
        self.DATA            = 'data'
        self.PREDICT         = 'predict'
        self.GENERATE        = 'generate'
        self.CONNECT         = 'connect'
        self.AGGREGATE       = 'aggregate'

    def _logging_setup(self, args):

        if args.verbosity not in range(1, 6):
            raise Exception("Logging verbosity must be a positive integer between 1 and 5.")

        logger = logging.getLogger('')
        logger.setLevel(debug[args.verbosity])
        log_format = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                       datefmt="%Y-%m-%d %H:%M:%S %p")

        stream_logger = logging.StreamHandler(sys.stdout)
        stream_logger.setFormatter(log_format)
        stream_logger.setLevel(debug[args.verbosity])
        logger.addHandler(stream_logger)

        if args.subparser_name!=self.DATA:
            file_logger = logging.FileHandler(os.path.join(args.output, args.log), 'a')
            file_logger.setFormatter(log_format)
            logger.addHandler(file_logger)

    def _check_general(self, args):
        '''
        Check general input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''

        dependencies = {'hmmsearch': "http://hmmer.org/download.html",
                        'diamond': "https://github.com/bbuchfink/diamond",
                        'R': "https://www.r-project.org",
                        'parallel': "https://www.gnu.org/software/parallel",
                        'prodigal': "https://github.com/hyattpd/Prodigal/wiki/installation",
                        'mmseqs': "https://github.com/soedinglab/MMseqs2"}
        
        missing_dependencies = list()
        
        for dependency in dependencies.keys():

            if shutil.which(dependency) == None:
                missing_dependencies.append(dependency)
        
        if len(missing_dependencies)>0:
            dependency_string = '\n'.join(['\t%s\t%s' % (d, dependencies[d]) for d in missing_dependencies])
            raise Exception('The following dependencies need to be installed to run enrichm:\n%s' % (dependency_string))

        # We dont need an output directory for the DATA pipeline
        if args.subparser_name!=self.DATA:
            # Set up working directory
            if not args.output:
                args.output = '%s-enrichm_%s_output' % (time.strftime("%Y-%m-%d_%H-%M"), args.subparser_name)

            if(os.path.isdir(args.output) or os.path.isfile(args.output)):

                if args.force:

                    if os.path.isdir(args.output):
                        shutil.rmtree(args.output)

                    else:
                        os.remove(args.output)

                else:
                    raise Exception("File '%s' exists." % args.output)

            os.mkdir(args.output)   

    def _check_annotate(self, args):
        '''
        Check annotate input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        if args.cut_ko:
            if int(Data.CURRENT_VERSION.split('_')[-1].replace('v', '')) < 10:
                raise Exception("EnrichM database needs to be version 10 or higher to use KO HMM cutoffs. Please run enrichm data.")

        # ensure either a list of genomes or a directory of genomes have been specified
        if not(args.genome_files or args.genome_directory or args.protein_directory or args.protein_files):
            raise Exception("Input error: Either a list of genomes or a directory of genomes need to be specified.")

        if len([x for x in [args.genome_files, args.genome_directory, args.protein_directory, args.protein_files] if x]) != 1:
            raise Exception("Input error: Only one type of input can be specified (--genome_files, --genome_directory, --protein_directory, or --protein_files).")
        
        if not args.suffix:
            
            if(args.genome_directory or args.genome_files):
                args.suffix = '.fna'
            
            elif(args.protein_directory or args.protein_files):
                args.suffix = '.faa'
        
        if(args.id>1 or args.id<0):
            raise Exception("Identity (--id) must be between 0 and 1.")
        
        if(args.aln_query>1 or args.aln_query<0):
            raise Exception("Alignment to query cutoff (--aln_query) must be between 0 and 1")
        
        if(args.aln_reference>1 or args.aln_reference<0):
            raise Exception("Alignment to reference cutoff (--aln_reference) must be between 0 and 1")
        
        if any([args.cut_ga, args.cut_nc, args.cut_tc]):
            if len([x for x in [args.cut_ga, args.cut_nc, args.cut_tc] if x])>1:
                raise Exception("Only one of the following can be selected: --cut_ga, --cut_nc, --cut_tc")
            
            if args.evalue:
                logging.warning('selecting one of the following overrides evalue thresholds: --cut_ga, --cut_nc, --cut_tc')
            
    def _check_enrichment(self, args):
        '''
        Check enrichment input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object

        Output
        ------
        '''
        ### ~ TODO: Check Multi test correction inputs...
        types = [args.ko, args.pfam, args.tigrfam, args.hypothetical, args.cazy, args.ec, args.ko_hmm]
        
        if not any(types):
            
            raise Exception("Input Error: One of the following flags must be specified: --ko --pfam --tigrfam --hypothetical --cazy")
        
        if len([x for x in types if x])>1:
            
            raise Exception("Only one of the following flags may be specified: --ko --pfam --tigrfam --hypothetical --cazy")

        if not args.abundance and args.abundance_metadata:
           raise Exception("values for both --abundance and --abundance_metadata are required") 

    def _check_classify(self, args):  
        '''
        Check classify input and output options are valid.
        
        Parameters
        ----------
        args    - object. Argparse object
        '''
        # Ensure either an annotation matrix or list file has been specified:
        if not(args.genome_and_annotation_file or args.genome_and_annotation_matrix):
            raise Exception("Input error: An input file must be specified to either \
--genome_and_annotation_file or --genome_and_annotation_matrix")

        elif(args.aggregate and args.genome_and_annotation_file):
            raise Exception("--aggregate needs to be run with the genome and annotation matrix")

    def _check_build(self, args):
        '''
        Check build input and output options are valid.
        
        Parameters
        ----------
        args    - object. Argparse object
        '''
        if not(args.metadata or args.abundances):
            
            raise Exception("No metadata or abundance information provided to build.")

    def _check_network(self, args):
        '''
        Check network (explore, pathway) input and output options are valid.
        
        Parameters
        ----------
        args    - object. Argparse object
        '''

        if any([args.abundance, args.abundance_metadata]):
            if not (args.abundance and args.abundance_metadata):
                raise Exception("Both abundance and abundance metadata need to be specified")

        if args.subparser_name==NetworkAnalyser.PATHWAY:
            args.depth              = None
            args.queries            = None
            args.starting_compounds = None
            args.steps              = None
            args.number_of_queries  = None
        
        if args.subparser_name==NetworkAnalyser.TRAVERSE:
            args.depth              = None
            args.queries            = None
    
        if args.subparser_name==NetworkAnalyser.EXPLORE:
            args.filter             = None  
            args.limit              = None 
            args.starting_compounds = None
            args.steps              = None
            args.number_of_queries  = None

            if not(args.queries):

                if args.depth:
                    logging.warning("--depth argument ignored without --queries flag")

    def _check_predict(self, args):
        '''
        Inputs
        ------
        
        Outputs
        -------
        '''
        pass

    def _check_generate(self, args):
        '''
        Parameters
        ----------
        
        Output
        ------
        '''
        pass

    def _check_connect(self, args):
        '''
        Check connect
        
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        if(args.cutoff > 1.0 and args.cutoff < 0.0):
           
           raise Exception("Cutoff needs to be between 0 - 1") 
    
    def main(self, args, command):
        '''
        Parameters
        ----------
        
        Output
        ------
        '''
        self._check_general(args)
        self._logging_setup(args)

        logging.info("Running command: %s" % ' '.join(command))

        if args.subparser_name == self.DATA:
            d = Data()
            d.do(args.uninstall)
        
        if args.subparser_name == self.ANNOTATE:
            self._check_annotate(args)
            a = Annotate(# Define inputs and outputs
                         args.output,
                         # Define type of annotation to be carried out
                         args.ko,
                         args.ko_hmm,
                         args.pfam,
                         args.tigrfam,
                         args.clusters,
                         args.orthologs,
                         args.cazy,
                         args.ec,
                         # Cutoffs
                         args.evalue,
                         args.bit,
                         args.id,
                         args.aln_query, 
                         args.aln_reference, 
                         args.c,
                         args.cut_ga,
                         args.cut_nc,
                         args.cut_tc,
                         args.cut_ko,
                         args.inflation,
                         args.chunk_number,
                         args.chunk_max,
                         args.count_domains,
                         # Parameters
                         args.threads,
                         args.parallel,
                         args.suffix,
                         args.light)
            a.do(args.genome_directory,
                 args.protein_directory, 
                 args.genome_files,
                 args.protein_files)

        
        elif args.subparser_name == self.CLASSIFY:
            self._check_classify(args)
            c = Classify()
            c.do(args.custom_modules, 
                 args.cutoff,
                 args.aggregate,
                 args.genome_and_annotation_file,
                 args.genome_and_annotation_matrix,
                 args.output)

        elif args.subparser_name == self.ENRICHMENT: 
            self._check_enrichment(args)
            e = Enrichment()
            e.do(# Input options
                 args.annotate_output,
                 args.metadata,
                 args.abundance,
                 args.abundance_metadata,
                 # Runtime options
                 args.genomes_to_compare_with_group,
                 args.pval_cutoff,
                 args.proportions_cutoff,
                 args.threshold,
                 args.multi_test_correction,
                 args.batchfile,
                 args.processes,
                 args.ko,
                 args.pfam,
                 args.tigrfam,
                 args.hypothetical,
                 args.cazy,
                 args.ec,
                 args.ko_hmm,
                 # Outputs
                 args.output)

        elif args.subparser_name == self.CONNECT:
            self._check_connect(args)
            c = Connect()
            c.do(args.annotate_output,
                 args.metadata,
                 args.custom_modules,
                 args.cutoff,
                 args.output)

        elif args.subparser_name == NetworkAnalyser.PATHWAY:
            self._check_network(args)
            na=NetworkAnalyser(args.metadata)
            na.do(args.matrix,
                  args.tpm_values,
                  args.abundance,
                  args.abundance_metadata,
                  args.metabolome,
                  args.enrichment_output,
                  args.depth,
                  args.filter,
                  args.limit,
                  args.queries,
                  args.starting_compounds, 
                  args.steps,
                  args.number_of_queries,
                  args.output)
        
        if args.subparser_name == self.PREDICT:
            self._check_predict(args)
            p = Predict()
            p.do(args.forester_model_directory,
                 args.input_matrix,
                 args.output)

        elif args.subparser_name == self.GENERATE:
            self._check_generate(args)
            gm = GenerateModel()
            gm.do(args.input_matrix,
                  args.groups,
                  args.model_type,
                  args.testing_portion,
                  args.grid_search,
                  args.threads,
                  args.output)
        
        logging.info('Done!')



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import logging
import os
import numpy as np
from scipy import stats
from itertools import combinations
# Local
from enrichm.genome import Genome, AnnotationParser
###############################################################################

class Compare:

	def __init__(self):

		# Result types
		self.gc = "GC content"
		self.length = "Genome length"
		self.tests = "Tests"

	def _pan_genome(self, genome_dict, metadata):
		result = dict()
		for group, genomes in metadata.items():
			group_genomes = [genome_dict[genome] for genome in genomes]
			group_core_genome_size = len(set.intersection(*[genome.clusters for genome in group_genomes]))
			group_genome_lengths = [len(genome.sequences) for genome in group_genomes]
			group_genome_mean = float(sum(group_genome_lengths)) / float(len(group_genome_lengths))
			group_core_size = round((float(group_core_genome_size) / group_genome_mean)*100, 2)
			result[group] = [group_core_genome_size, group_core_size]
		return result

	def gather_stats(self, number_list):
		number_array = np.array(number_list)
		mean = number_array.mean()
		median = np.median(number_array)
		std = np.std(number_array)
		minimum = min(number_array)
		maximum = max(number_array)
		p90 = np.percentile(number_array, 90)
		p10 = np.percentile(number_array, 10)

		return [mean, median, std, minimum, maximum, p90, p10]

	def run_mannwhitneyu(self, feature, feature_list):
		
		test_results = dict()

		for combination in combinations(feature_list.keys(), 2):

			group_1 = np.array(feature_list[combination[0]])
			group_2 = np.array(feature_list[combination[1]])

			test_results[combination] \
				= stats.mannwhitneyu(group_1, group_2)
		return test_results

	def _compare_features(self, genome_dict, metadata):
		# GC, size, coding density
		keys = list(metadata.keys())+[self.tests]

		results = {self.gc:		{x:[] for x in keys},
				   self.length:	{x:[] for x in keys}}

		for group, genomes in metadata.items():
			group_genomes = [genome_dict[genome] for genome in genomes]
			group_lengths = list()
			group_gc = list()
			
			for genome in group_genomes:
				group_gc.append(genome.gc)
				group_lengths.append(genome.length)

			results[self.gc][group].append([group_gc, self.gather_stats(group_gc)]) 
			results[self.length][group].append([group_lengths, self.gather_stats(group_lengths)]) 

		for feature, feature_list in results.items():
			test_result = self.run_mannwhitneyu(feature, feature_list)
			results[self.tests].append([feature, test_result])
		
		return results
	
	def saturate(self, genomes_list, orthologs, output):
		result = dict()

		for group in range(1, len(genomes_list)+1):
			
			core_counts = list()
			accessory_counts = list()

			for combination in combinations(genomes_list, group):

				core_count = 0
				accessory_count = 0

				combination_length = len(combination)

				for ortholog in orthologs:
					hits = len([True for genome in combination if ortholog in genome.orthologs])
					if hits==combination_length:
						core_count+=1
					else:
						accessory_count+=1
				core_counts.append(core_count)
				accessory_counts.append(accessory_count)

			result[group] = {"core":self.gather_stats(core_counts), 'accessory':self.gather_stats(accessory_counts)}

		with open(output, 'w') as out_io:
			out_io.write('\t'.join(["Group size",
									"Core mean",
									"Core median",
									"Core standard deviation",
									"Core minimum",
									"Core maximum",
									"Core p90",
									"Core p10",
									"Accessory mean",
									"Accessory median",
									"Accessory standard deviation",
									"Accessory minimum",
									"Accessory maximum",
									"Accessory p90",
									"Accessory p10"]) +'\n')
			for group, stats in result.items():
				output_line = [str(group)]
				output_line += [str(x) for x in stats["core"]]
				output_line += [str(x) for x in stats["accessory"]]
				out_io.write('\t'.join(output_line) +'\n')

	def write_pan_genome_results(self, results, output_path):
		header = ['Genome', "Core genome size", "Percent of average genome size"]
		with open(output_path, 'w') as out_io:
			out_io.write('\t'.join(header) + '\n')
			for genome, results_list in results.items():
				output_line = [genome] + [str(x) for x in results_list]
				out_io.write('\t'.join(output_line) + '\n')

	def do(self, genome_list, metadata, output_directory):
		'''
		Parameters
		----------
		genome_objects - List.
		'''
		genome_dict = {genome.name:genome for genome in genome_list}
		ortholog_list = set.union(*[genome.orthologs for genome in genome_list])

		for group_name, genome_list in metadata.items():	
			logging.info('Generating pan genome saturation curve for group: %s' % group_name)
			group_list = [genome_dict[g] for g in genome_list]

			output = os.path.join(output_directory, "%s_saturation.tsv" % (group_name))
			self.saturate(group_list, ortholog_list, output)

		pan_genome_results = self._pan_genome(genome_dict, metadata)
		self.write_pan_genome_results(pan_genome_results, os.path.join(output_directory, "core_genome_size.tsv"))
		#if hasattr(list(genome_dict.values())[0], "gc"):
		#	compare_features_results = self._compare_features(genome_dict, metadata)
		#	self.write_compare_features_results(compare_features_results)
		
